%% DT_T2_d2.m
% T-61.2010 Datasta tietoon 
% Tietokoneharjoitus #2
% Demo 2
%
% Jukka Parviainen, jukka.parviainen@aalto.fi, 20111026

%% Pkomponenttianalyysi / Principal Component Analysis
%
% Tss demossa 2-ulotteinen datamatriisi X.
% Vain 2D, jotta voidaan visualoida helposti xy-tasossa.
% 
% - piirretn alkuperinen data
% - tehdn PCA: 
%   - keskiarvon poistaminen
%   - kovarianssimatriisi
%   - tmn ominaisarvot
% - piirretn uudet PCA-akselit (siirto + kierto), jotka
%    ovat edelleen ortogonaalisia (90 aste kulma)
% - projisoidaan PCA1:lle (dimension vhentminen)
% - todetaan dekorrelaatio

%% Muutamia puhdistuskomentoja
clear      % poistaa kaikki muuttujat tytilasta
shg        % nostaa viimeisimmn piirtoikkunan yls
clf        % puhdistaa aktiivisen piirtoikkunan
close all  % sulkee kaikki ikkunat
clc        % puhdistaa komentoikkunan tekstist

%% Data
% Valitse joko 6 ihmisen pituus ja paino tai
%  tasajakautunutta (rand) dataa tai
%  normaalijakautunutta (randn) dataa.
% Valitse joko kommentoimalla kaksi muuta pois
%  tai aja kaikki lpi, jolloin viimeisen oleva
%  "j voimaan".
% gridIx ja gridIy tarvitaan piirroskomennoissa myhemmin.

% valinta 1:
%X = [184 175 162 191 170 179; ...
%    76   70   52  78  75  72];
%gridIx = 10;
%gridIy = 10;

% valinta 2:
%N  = 20; 
%dummy = randn(1,N)+4;
%X = [dummy; 2*dummy-3+0.3*randn(1,N)];

% valinta 3:
N  = 20;
dummy = rand(1,N)+2;
X = [dummy; -3*dummy+0.5*rand(1,N)];

dim  = size(X, 1);  % 'variables', muuttujien lkm
N    = size(X, 2);  % 'observations', havaintojen lkm

disp(['Alkuperinen datamatriisi X (' num2str(dim) ' x ' num2str(N) ')']);
disp(X);

%% Piirretn (x1, x2)-koordinaatistoon

%xTickMin = min(X(1,:)) - 0.1*abs(min(X(1,:))); % hmmm
xMinAxis = min(X(1,:)) - 0.1*(max(X(1,:)) - min(X(1,:)));
xMaxAxis = max(X(1,:)) + 0.1*(max(X(1,:)) - min(X(1,:)));
yMinAxis = min(X(2,:)) - 0.1*(max(X(2,:)) - min(X(2,:)));
yMaxAxis = max(X(2,:)) + 0.1*(max(X(2,:)) - min(X(2,:)));

figure(221); clf;
p1 = plot(X(1,:), X(2,:), 'bo');
set(p1, 'MarkerFaceColor',[0 0 1]);
grid on;
%axis equal  % hmmm
%set(gca,'XTick',[150:10:200]); % hmmm
%set(gca,'YTick',[50:10:100]);  % hmmm
%axis([140 205 45 95]); % hmmm
title('Datapisteet');

intervalForTexts = round(size(X,2)/5)+1;
for k = [1 : intervalForTexts : size(X, 2)]
    text(X(1,k),X(2,k),['#' num2str(k) ':(' num2str(X(1,k)) ',' num2str(X(2,k)) ')']);
end;

%axis equal;
axis([xMinAxis xMaxAxis yMinAxis yMaxAxis]);

%% Datan keskiarvoistaminen (vaihe 1)

ka = (mean(X,2));                    % laske keskiarvo (vaihe 1)
XK = X - repmat(ka, 1, size(X,2));   % poista keskiarvo (vaihe 1)

disp('Keskiarvo ka:');
disp(ka);
disp('joka poistetaan ja saadaan XK:');
disp(XK);

figure(221); hold on;
XC = ka(1)+[-100 0; 100 0];
YC = ka(2)+[0 -100; 0 100];
lin8 = line(XC, YC);
set(lin8,'Color',[0 1 0]);
title(['Keskiarvoistetut datapisteet, ka = (' num2str(ka(1)) ',' num2str(ka(2)) ')']);

for k = [1 : intervalForTexts : size(X, 2)]
    t1 = text(X(1,k)+0.2,X(2,k)+0.2,['#' num2str(k) ':(' num2str(XK(1,k)) ',' num2str(XK(2,k)) ')']);
    set(t1,'Color',[0 1 0]);
end;

axis([xMinAxis xMaxAxis yMinAxis yMaxAxis]);

%% Etsi pkomponentit
% Tehtvn etsi ensin siis se suunta, jossa
% datan variaatio on suurin.

%% Muodostetaan kovarianssimatriisi (vaihe 2)
% Kovarianssimatriisin kooksi tulee (d x d) eli tss 
% tapauksessa (2 x 2).
% Kerroin 1/(n-1) tuottaa harhattoman kovarianssimatriisin.

C = (1/size(XK,2))*XK*XK';           % laske C (vaihe 2)
disp('Kovarianssimatriisi C:');
disp(C);

%% Lasketaan ominaisarvot (vaihe 3)
% Ominaisarvot ja -vektorit lasketaan nelimatriisista.
% Ratkaistaan siis yhtl:
%
% $$ A x = \lambda x $$
%
% Matlabissa: lue "help eig"
%
% Ominaisvektorit V:n sarakkeita.
% Vastaavat ominaisarvot D:n diagonaalilla.

[V, D] = eig(C);                     % lasketaan om.vekt. (vaihe 3)

%% Jrjestetn ominaisvektorit ominaisarvojen suuruusjrjestykseen
%
% Jrjestmisen jlkeen ollaan varmoja, ett V:n ensimmisess sarakkeessa
% on ominaisvektori, jota vastaava ominaisarvo on suurin.
% Mys muut vektorit ovat ominaisarvojen suuruusjrjestyksess.

diagD = diag(D);                      % (vaihe 3B) Matlabista johtuen knnetn
[dummy1, ind] = sort(diagD);
ind = flipud(ind);
V     = V(:, ind);
D     = D(:, ind);
diagD = diagD(ind);

disp('C:n ominaisvektorit sarakkeissa:');
disp(V);

disp('C:n ominaisarvot vastaavasti suuruusjrjestyksess:');
disp(diagD');

%% Ominaisarvot
%
% Lasketaan kumulatiivinen summa ominaisarvoista.
%
% Tllin ensimmiset m ominaisarvoa selitt f(m) prosenttia
% koko datan vaihtelusta. Tt voidaan kytt dimension
% pudotukseen, koska useimmiten "viimeisten" dimensioiden
% tuottama varianssi on varsin pient.

lambda = sum(D);
lambdacs = cumsum(lambda);
lambdasum = sum(lambda);

disp('Ominaisarvo #k | Kumul. summa ominaisarvoista %');
disp([ [1:length(lambda)]'  100*lambdacs'/lambdasum]);

%% Projisoidaan 2D-data 1D-avaruuteen (vaihe projektio)
%
% Kytetn projektiomatriisina W "suurinta" 
% pkomponenttivektoria.
%
% Datan XK projektio suoralle (hypertasolle) W:
%
% $$ PK = W' * XK $$
%
% Dimensiotarkastelu:
%  [W]  = (2x1)
%  [XK] = (2x20)    % N==20 kpl 2-ulotteisia
%  [PK] = [W' * XK] = (1x2) * (2x20) = (1x20)  % 20 kpl 1-ulott.

disp('PCA1-vektori:');
disp(V(:,1))
disp('Datan varianssista selittyy (%):');
disp(100*sum(D(:,1))/sum(D(:)));

W  = V(:,1);       % projektiomatriisiksi suurin pkomponentti
PK = W' * XK;      % (vaihe projektio)
disp('Projektiopisteiden arvot 1D:ss:');
disp(PK);

%% Piirretn projektiopisteet viel 2D-avaruuteen
%
% Projektipisteet PK ovat nyt 1-ulotteisia eli sijaitsevat
% suoralla, jonka suurin ominaisvektori W mr.
%
% Jos ajatellaan dekoodaamisena, niin meill on
% N==20 kpl 1D-pisteit (20 yksikk) ja tarvittavat
% ominaisvektorit (2 yksikk) sen sijaan ett 
% meill olisi alkuperiset 20 kpl 2D-pisteit (40 yksikk).
%
% Takaisinpin
%
% $$ \hat{XK} = W * PK $$
%
% eli nyt XKhattu on XK:n projektio alkuperisess 2D:ss

%XKhattu = W * PK + repmat(ka, 1, size(PK,2));
%XP      = XKHattu(1,:);
%YP      = XKHattu(2,:);
XP = PK * V(1,1) + ka(1);
YP = PK * V(2,1) + ka(2);
p2 = plot(XP, YP, 'ro', 'MarkerSize', 12);

% piirroksia:
for k = [1 : intervalForTexts : size(XP, 2)]
    t1 = text(XP(k)+0.2,YP(k)-0.2,['#' num2str(k) ':(' num2str(XP(k)) ',' num2str(YP(k)) ')']);
    set(t1,'Color',[1 0 0]);
end;

p3 = line([-V(1,1) V(1,1)]*100+ka(1), [-V(2,1) V(2,1)]*100+ka(2));
set(p3, 'LineStyle', '-.', 'Color', [1 0 0]);

p4 = line([-V(1,2) V(1,2)]*100+ka(1), [-V(2,2) V(2,2)]*100+ka(2));
set(p4, 'LineStyle', ':', 'Color', [1 0 0]);

disp('Projektiopisteiden arvot 2D:ss:');
disp([XP; YP]);

legend({'Datapiste','X-suunta','Y-suunta','Projektiopiste PCA1:lle','PCA1-suunta','PCA2-suunta'}, ... 
    'Location','SouthWest');

axis equal  % jos tt ei ole, niin PCA-akselit "eivt nyt olevan kohtisuorassa"
if (range(X(1,:)) < range(X(2,:)))
    set(gca,'YLim',[min(X(2,:)) max(X(2,:))]);
else
    set(gca,'XLim',[min(X(1,:)) max(X(1,:))]);
end;

%% Dekorrelointi
%
% Tss dimensiotarkastelu
%  [XYP] = [V' * XK] = (2x2) * (2x20) = (2x20)
% eli ei pudoteta dimensiota vaan vain dekorreloidaan
% eli irti lineaarisista riippuvuuksista.

XYP = V' * XK;
figure(222); clf;
plot(XYP(1,:), XYP(2,:), 'o');
grid on;
title('Dekorreloidut pisteet eli kuvattu uudessa PCA-koordinaatiossa');
xlabel('PCA1');
ylabel('PCA2');

for k = [1 : intervalForTexts : size(XYP, 2)]
    t1 = text(XYP(1,k),XYP(2,k),['#' num2str(k) ':(' num2str(XYP(1,k)) ',' num2str(XYP(2,k)) ')']);
    set(t1,'Color',[1 0 0]);
end;

% Lasketaan korrelaatiokertoimet alkuperiselle ja dekorreloidulle datalle:
%
% Korrelaatiomatriisi (dim x dim) on symmetrinen. Diagonaalilla on
%  aina ykksi, koska muuttujan korrelaatio itseens on 1. 
%  Korrelaatiomatriisi kuvaa muuttujien
%  vlist lineaarista riippuvuutta muuttujapareittain. 
%  Luku lhell ykkst vastaa
%  positiiviista korrelaatiota (kun x kasvaa niin y kasvaa),
%  -1 negatiivista (kun x kasvaa niin y vhenee) ja arvo 0
%  ettei x:ll ja y:ll ole lineaarista korrelaatiota.
% 
% Huomaa, ett x:ll ja y:ll voi olla funktionaalinen yhteys vaikka
%  korrelaatio olisi nolla.
%
%          [ r(1,1)   r(1,2) .. r(  1,dim) ]
%    R  =  [ r(2,1)   r(2,2) .. r(  2,dim) ]
%          [ ..       ..     .. ..         ]
%          [ r(dim,1) ..     .. r(dim,dim) ]
%
% jossa siis r(1,2)==r(2,1) jne ja r(k,k)==1.

R_XK  = corrcoef(XK');
R_XYP = corrcoef(XYP');

disp('Alkuperisen datan korrelaatiomatriisi:');
disp(R_XK);
disp('Dekorreloidun datan korrelaatiomatriisi:');
disp(R_XYP);

%% Valkaisu
% PCA + varianssien pakottaminen ykkseksi

